/* MemoServ functions.
 *
 * Services is copyright (c) 1996-1998 Andy Church.
 *     E-mail: <achurch@dragonfire.net>
 * This program is free but copyrighted software; see the file COPYING for
 * details.
 */

#include "services.h"

static MemoList *memolists[256];	/* One for each initial character */

const char s_MemoServ[] = MEMOSERV_NAME;

#include "ms-help.c"


static MemoList *find_memolist(const char *nick);
static void alpha_insert_memolist(MemoList *ml);
static void del_memolist(MemoList *ml);

static void do_help(const char *source);
static void do_send(const char *source);
static void do_list(const char *source);
static void do_read(const char *source);
static void do_del(const char *source);
static void do_set(const char *source);
static void do_set_notify(NickInfo *ni, char *param);
static void do_set_limit(NickInfo *ni, char *param);

/*************************************************************************/

/* Return information on memory use.  Assumes pointers are valid. */

void get_memoserv_stats(long *nrec, long *memuse)
{
    long count = 0, mem = 0;
    int i, j;
    MemoList *ml;

    for (i = 0; i < 256; i++) {
	for (ml = memolists[i]; ml; ml = ml->next) {
	    count++;
	    mem += sizeof(*ml);
	    mem += sizeof(Memo) * ml->n_memos;
	    for (j = 0; j < ml->n_memos; j++) {
		if (ml->memos[j].text)
		    mem += strlen(ml->memos[j].text)+1;
	    }
	}
    }
    *nrec = count;
    *memuse = mem;
}

/*************************************************************************/
/*************************************************************************/

/* memoserv:  Main MemoServ routine. */

void memoserv(const char *source, char *buf)
{
    char *cmd, *s;
    NickInfo *ni;

    cmd = strtok(buf, " ");

    if (!cmd) {
	return;

    } else if (stricmp(cmd, "\1PING") == 0) {
	if (!(s = strtok(NULL, "")))
	    s = "\1";
	notice(s_MemoServ, source, "\1PING %s", s);

#ifdef SKELETON

    } else {
	notice(s_MemoServ, source, "%s is currently offline.", s_MemoServ);
    }

#else

    } else if (stricmp(cmd, "HELP") == 0) {
	do_help(source);

    } else if (!(ni = findnick(source))) {
	notice(s_MemoServ, source, "Your nick is not registered.  Type "
			"\2/msg %s HELP\2 for information on registering "
			"your nickname.", s_NickServ);

    } else if (!(ni->flags & (NI_RECOGNIZED | NI_IDENTIFIED))) {
	notice(s_MemoServ, source,
		"Please identify with %s first, using the command:",
		s_NickServ);
	notice(s_MemoServ, source,
		"/msg %s IDENTIFY \37password\37", s_NickServ);

    } else if (stricmp(cmd, "SEND") == 0) {
	do_send(source);

    } else if (!(ni->flags & NI_IDENTIFIED)) {
	notice(s_MemoServ, source, "Identification required for that command."
			"  Type \2/msg %s IDENTIFY \37password\37\2 to "
			"access your memos.", s_NickServ);

    } else if (stricmp(cmd, "LIST") == 0) {
	do_list(source);

    } else if (stricmp(cmd, "READ") == 0) {
	do_read(source);

    } else if (stricmp(cmd, "DEL") == 0) {
	do_del(source);

    } else if (stricmp(cmd, "SET") == 0) {
	do_set(source);

#if 0	/* commented out because we don't have any admin-only commands */
    } else if (!is_services_admin(source)) {
	notice(s_MemoServ, source,
		"Unrecognized command \2%s\2.  Type \"/msg %s HELP\" for help.",
		cmd, s_MemoServ);
#endif

    } else {
	notice(s_MemoServ, source,
		"Unrecognized command \2%s\2.  Type \"/msg %s HELP\" for help.",
		cmd, s_MemoServ);

    }

#endif
}

/*************************************************************************/

#ifndef SKELETON


/* load_ms_dbase, save_ms_dbase:  Load/save memo data. */

void load_ms_dbase(void)
{
    FILE *f;
    int ver, i, j, len;
    MemoList *ml;
    Memo *memos;

    if (!(f = open_db(s_MemoServ, MEMOSERV_DB, "r")))
	return;
    switch (ver = get_file_version(f, MEMOSERV_DB)) {
      case 3:
      case 2:
      case 1:
	for (i = 33; i < 256; ++i) {
	    while (fgetc(f) == 1) {
		ml = smalloc(sizeof(MemoList));
		if (1 != fread(ml, sizeof(MemoList), 1, f))
		    fatal_perror("Read error on %s", MEMOSERV_DB);
		alpha_insert_memolist(ml);
		ml->memos = memos = smalloc(sizeof(Memo) * ml->n_memos);
		fread(memos, sizeof(Memo), ml->n_memos, f);
		for (j = 0; j < ml->n_memos; ++j, ++memos)
		    memos->text = read_string(f, MEMOSERV_DB);
	    }
	}
	break;
      default:
	fatal("Unsupported version number (%d) on %s", ver, MEMOSERV_DB);
    } /* switch (version) */
    fclose(f);
}

void save_ms_dbase(void)
{
    FILE *f;
    int i, j, len;
    MemoList *ml;
    Memo *memos;

    if (!(f = open_db(s_MemoServ, MEMOSERV_DB, "w")))
	return;
    for (i = 33; i < 256; ++i) {
	for (ml = memolists[i]; ml; ml = ml->next) {
	    fputc(1, f);
	    if (1 != fwrite(ml, sizeof(MemoList), 1, f) ||
		    ml->n_memos !=
			fwrite(ml->memos, sizeof(Memo), ml->n_memos, f))
		fatal_perror("Write error on %s", MEMOSERV_DB);
	    for (memos = ml->memos, j = 0; j < ml->n_memos; ++memos, ++j)
		write_string(memos->text, f, MEMOSERV_DB);
	}
	fputc(0, f);
    }
    fclose(f);
    remove(MEMOSERV_DB ".save");
}

/*************************************************************************/

/* check_memos:  See if the given nick has any waiting memos, and send a
 *               NOTICE to that nick if so.
 */

void check_memos(const char *nick)
{
    NickInfo *ni;
    MemoList *ml;

    if (!(ni = findnick(nick)) || !(ni->flags & NI_MEMO_SIGNON))
	return;
    if (ml = find_memolist(nick)) {
	notice(s_MemoServ, nick, "You have %d memo%s.", ml->n_memos,
		ml->n_memos == 1 ? "" : "s");
	notice(s_MemoServ, nick, "Type \2/msg %s %s\2 to %s.",
		s_MemoServ,
		ml->n_memos == 1 ? "READ 1" : "LIST",
		ml->n_memos == 1 ? "read it" : "list them");
	if (ni->memomax > 0 && ml->n_memos >= ni->memomax) {
	    notice(s_MemoServ, nick,
		"\2Warning:\2 You %s your maximum number of memos.  You "
		"will be unable to receive any new memos until you delete "
		"some of your current ones.",
		ml->n_memos > ni->memomax ? "are over" : "have reached");
	}
    }
}

/*************************************************************************/

/* clear_memos:  Clear all memos (if any) for a given user. */

void clear_memos(const char *nick)
{
    MemoList *ml;

    if (ml = find_memolist(nick))
	del_memolist(ml);
}

/*************************************************************************/
/*********************** MemoServ private routines ***********************/
/*************************************************************************/

/* find_memolist:  Find the memo list for a given nick.  Return NULL if
 *                 none.
 */

static MemoList *find_memolist(const char *nick)
{
    MemoList *ml;
    int i;

    for (ml = memolists[tolower(*nick)];
			ml && (i = stricmp(ml->nick, nick)) < 0;
			ml = ml->next)
	;
    return i==0 ? ml : NULL;
}

/*************************************************************************/

/* alpha_insert_memolist:  Insert a memo list alphabetically into the
 *                         database.
 */

static void alpha_insert_memolist(MemoList *ml)
{
    MemoList *ml2, *ml3;
    char *nick = ml->nick;

    for (ml3 = NULL, ml2 = memolists[tolower(*nick)];
			ml2 && stricmp(ml2->nick, nick) < 0;
			ml3 = ml2, ml2 = ml2->next)
	;
    ml->prev = ml3;
    ml->next = ml2;
    if (!ml3)
	memolists[tolower(*nick)] = ml;
    else
	ml3->next = ml;
    if (ml2)
	ml2->prev = ml;
}

/*************************************************************************/

/* del_memolist:  Remove a nick's memo list from the database.  Assumes
 *                that the memo count for the nick is non-zero.
 */

static void del_memolist(MemoList *ml)
{
    int i;

    if (ml->next)
	ml->next->prev = ml->prev;
    if (ml->prev)
	ml->prev->next = ml->next;
    else
	memolists[tolower(*ml->nick)] = ml->next;
    free(ml->memos);
    free(ml);
}

/*************************************************************************/
/*********************** MemoServ command routines ***********************/
/*************************************************************************/

/* Return a help message. */

static void do_help(const char *source)
{
    char *cmd = strtok(NULL, "");
    char buf[BUFSIZE];

    if (cmd && strnicmp(cmd, "SET", 3) == 0 && isspace(cmd[3])) {
	char *s = strspn(cmd+4, " ") + cmd+4;
	if (stricmp(cmd, "LIMIT") == 0 && is_services_admin(source)) {
	    notice_list(s_MemoServ, source, set_limit_help);
	    return;
	}
    }
    snprintf(buf, sizeof(buf), "%s%s", s_MemoServ, cmd ? " " : "");
    strscpy(buf+strlen(buf), cmd ? cmd : "", sizeof(buf)-strlen(buf));
    helpserv(s_MemoServ, source, buf);
    if (cmd && stricmp(cmd, "SET") == 0 && is_services_admin(source))
	notice_list(s_MemoServ, source, servadmin_set_help);
}

/*************************************************************************/

/* Send a memo to a nick. */

static void do_send(const char *source)
{
    NickInfo *ni;
    MemoList *ml;
    Memo *m;
    char *nick = strtok(NULL, " ");
    char *text = strtok(NULL, "");
    User *u;
    time_t now = time(NULL);

    if (!text) {
	notice(s_MemoServ, source,
		"Syntax: \2SEND \37nick\37 \37memo-text\37\2");
	notice(s_MemoServ, source,
		"\2/msg %s HELP SEND\2 for more information.", s_MemoServ);

    } else if (!(ni = findnick(nick))) {
	notice(s_MemoServ, source, "Nick %s isn't registered.", nick);

    } else if ((u = finduser(source)) && u->lastmemosend+MEMO_SEND_WAIT > now) {
	u->lastmemosend = now;
	notice(s_MemoServ, source,
		"Please wait %d seconds before using the SEND command again.",
		MEMO_SEND_WAIT);

    } else {
	u->lastmemosend = now;
	ml = find_memolist(nick);
	if (!ml) {
	    ml = scalloc(sizeof(MemoList), 1);
	    strscpy(ml->nick, nick, NICKMAX);
	    alpha_insert_memolist(ml);
	}
	if (ni->memomax > 0 && ml->n_memos >= ni->memomax) {
	    notice(s_MemoServ, source,
		"%s currently has too many memos and cannot receive more.",
		nick);
	    return;
	}
	ml->n_memos++;
	ml->memos = srealloc(ml->memos, sizeof(Memo) * ml->n_memos);
	m = &ml->memos[ml->n_memos-1];
	strscpy(m->sender, source, NICKMAX);
	if (ml->n_memos > 1) {
	    m->number = m[-1].number + 1;
	    if (m->number < 1) {
		int i;
		for (i = 0; i < ml->n_memos; ++i)
		    ml->memos[i].number = i+1;
	    }
	} else
	    ml->memos[ml->n_memos-1].number = 1;
	m->time = time(NULL);
	m->text = sstrdup(text);
	notice(s_MemoServ, source, "Memo sent to %s.", nick);
	if ((ni->flags & NI_MEMO_RECEIVE) && finduser(nick)) {
	    notice(s_MemoServ, nick, "You have a new memo (#%d) from %s.",
			m->number, source);
	    notice(s_MemoServ, nick, "Type \2/msg %s READ %d\2 to read it.",
			s_MemoServ, m->number);
	}
    }
}

/*************************************************************************/

/* List the memos (if any) for the source nick. */

static void do_list(const char *source)
{
    MemoList *ml;
    Memo *m;
    int i;
    char timebuf[64];
    struct tm tm;

    if (!(ml = find_memolist(source))) {
	notice(s_MemoServ, source, "You have no memos.");

    } else {
	notice(s_MemoServ, source,
		"Memos for %s.  To read, type \2/msg %s READ \37num\37\2",
		source, s_MemoServ);
	notice(s_MemoServ, source, "Num  Sender            Date/Time");
	for (i = 0, m = ml->memos; i < ml->n_memos; ++i, ++m) {
	    tm = *localtime(&m->time);
	    strftime(timebuf, sizeof(timebuf), "%a %b %d %H:%M:%S %Y", &tm);
	    timebuf[sizeof(timebuf)-1] = 0;	/* just in case */
	    notice(s_MemoServ, source, "%3d  %-16s  %s %s",
			m->number, m->sender, timebuf, time_zone);
	}

    }
}

/*************************************************************************/

/* Read a memo. */

static void do_read(const char *source)
{
    MemoList *ml;
    Memo *m;
    char *numstr = strtok(NULL, " ");
    int num;

    if (!numstr || (num = atoi(numstr)) <= 0) {
	notice(s_MemoServ, source, "Syntax: \2READ \37num\37");
	notice(s_MemoServ, source,
		"\2/msg %s HELP READ\2 for more information.", s_MemoServ);

    } else if (!(ml = find_memolist(source))) {
	notice(s_MemoServ, source, "You have no memos.");

    } else {
	int i;
	for (i = 0; i < ml->n_memos; ++i) {
	    if (ml->memos[i].number == num)
		break;
	}
	if (i >= ml->n_memos) {
	    notice(s_MemoServ, source, "Memo %d does not exist!", num);
	} else {
	    m = &ml->memos[i];
	    notice(s_MemoServ, source,
		"Memo %d from %s.  To delete, type: \2/msg %s DEL %d\2",
		m->number, m->sender, s_MemoServ, m->number);
	    notice(s_MemoServ, source, "%s", m->text);
	}

    }
}

/*************************************************************************/

/* Delete a memo. */

static void do_del(const char *source)
{
    MemoList *ml;
    char *numstr = strtok(NULL, " ");
    int num, i;

    if (!numstr ||
	    ((num = atoi(numstr)) <= 0 && stricmp(numstr, "ALL") != 0)) {
	notice(s_MemoServ, source, "Syntax: \2DEL {\37num\37 | ALL}");
	notice(s_MemoServ, source,
		"\2/msg %s HELP DEL\2 for more information.", s_MemoServ);

    } else if (!(ml = find_memolist(source))) {
	notice(s_MemoServ, source, "You have no memos.");

    } else {
	if (num > 0) {
	    /* Delete a specific memo. */
	    for (i = 0; i < ml->n_memos; ++i) {
		if (ml->memos[i].number == num)
		    break;
	    }
	    if (i < ml->n_memos) {
		free(ml->memos[i].text); /* Deallocate memo text memory */
		--ml->n_memos;		 /* One less memo now */
		if (i < ml->n_memos)	 /* Move remaining memos down a slot */
		    bcopy(ml->memos + i+1, ml->memos + i,
					sizeof(Memo) * (ml->n_memos - i));
		notice(s_MemoServ, source, "Memo %d has been deleted.", num);
	    } else {
		notice(s_MemoServ, source, "Memo %d does not exist!", num);
	    }
	} else {
	    /* Delete all memos.  This requires freeing the memory holding
	     * the text of each memo and flagging that there are no memos
	     * left. */
	    for (i = 0; i < ml->n_memos; ++i)
		free(ml->memos[i].text);
	    ml->n_memos = 0;
	    notice(s_MemoServ, source, "All of your memos have been deleted.");
	}

	/* Did we delete the last memo?  If so, delete this MemoList. */
	if (ml->n_memos == 0)
	    del_memolist(ml);
    }
}

/*************************************************************************/

static void do_set(const char *source)
{
    char *cmd    = strtok(NULL, " ");
    char *param  = strtok(NULL, "");
    NickInfo *ni = findnick(source);
    User *u;

    if (readonly) {
	notice(s_MemoServ, source,
		"Sorry, memo option setting is temporarily disabled.");
	return;
    }

    if (!param) {

	notice(s_MemoServ, source,
		"Syntax: \2SET \37option\37 \37parameters\37\2");
	notice(s_MemoServ, source,
		"\2/msg %s HELP SET\2 for more information.", s_MemoServ);

    } else if (!(ni = findnick(source))) {

	notice(s_MemoServ, source, "Your nickname is not registered.");

    } else if (!(u = finduser(source)) || !(ni->flags & NI_IDENTIFIED)) {

	notice(s_MemoServ, source,
		"Password authentication required for that command.");
	notice(s_MemoServ, source,
		"Retry after typing \2/msg %s IDENTIFY \37password\37.",
		s_NickServ);
	if (!u)
	    log("%s: SET from nonexistent user %s", s_MemoServ, source);

    } else if (stricmp(cmd, "NOTIFY") == 0) {

	do_set_notify(ni, param);

    } else if (!is_services_admin(source)) {

	notice(s_MemoServ, source,
		"Unknown SET option \2%s\2.", strupper(cmd));
	notice(s_MemoServ, source,
		"\2/msg %s HELP SET\2 for more information.", s_MemoServ);

    } else if (stricmp(cmd, "LIMIT") == 0) {

	do_set_limit(ni, param);

    } else {

	notice(s_MemoServ, source,
		"Unknown SET option \2%s\2.", strupper(cmd));
	notice(s_MemoServ, source,
		"\2/msg %s HELP SET\2 for more information.", s_MemoServ);

    }
}

/*************************************************************************/

static void do_set_notify(NickInfo *ni, char *param)
{
    char *source = ni->nick;

    if (stricmp(param, "ON") == 0) {

	ni->flags |= NI_MEMO_SIGNON | NI_MEMO_RECEIVE;
	notice(s_MemoServ, source,
		"%s will now notify you of memos when you log on and when "
		"they are sent to you.", s_MemoServ);

    } else if (stricmp(param, "LOGON") == 0) {

	ni->flags |= NI_MEMO_SIGNON;
	ni->flags &= ~NI_MEMO_RECEIVE;
	notice(s_MemoServ, source,
		"%s will now notify you of memos when you log on or unset "
		"/AWAY.", s_MemoServ);

    } else if (stricmp(param, "NEW") == 0) {

	ni->flags &= ~NI_MEMO_SIGNON;
	ni->flags |= NI_MEMO_RECEIVE;
	notice(s_MemoServ, source,
		"%s will now notify you of memos when they are sent to you.",
		s_MemoServ);

    } else if (stricmp(param, "OFF") == 0) {

	ni->flags &= ~(NI_MEMO_SIGNON | NI_MEMO_RECEIVE);
	notice(s_MemoServ, source,
		"%s will not send you any notification of memos.", s_MemoServ);

    } else {

	notice(s_MemoServ, source, "Syntax: \2SET NOTIFY {ON|LOGON|NEW|OFF}\2");
	notice(s_MemoServ, source,
		"\2/msg %s HELP SET NOTIFY\2 for more information.",s_MemoServ);
    }
}

/*************************************************************************/

static void do_set_limit(NickInfo *ni, char *param)
{
    char *source = ni->nick;
    char *user = strtok(param, " ");
    char *s = strtok(NULL, " ");
    int limit;

    if (!user || !s || (limit = strtol(s, &s, 10), s && *s)) {

	notice(s_MemoServ, source,
		"Syntax: \2SET LIMIT \37user\37 \37limit\37\2");

    } else if (!(ni = findnick(user))) {

	notice(s_MemoServ, source, "Nickname \2%s\2 not registered!", user);

    } else {

	ni->memomax = limit;
	if (limit) {
	    notice(s_MemoServ, source,
			"Memo limit for \2%s\2 set to \2%d\2.", user, limit);
	} else {
	    notice(s_MemoServ, source,
			"Memo limit disabled for \2%s\2.", user);
	}

    }
}

/*************************************************************************/

#endif	/* !SKELETON */
